//  BMPx - The Dumb Music Player
//  Copyright (C) 2005-2006 BMPx development team.
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//  --
//
//  The BMPx project hereby grants permission for non GPL-compatible GStreamer
//  plugins to be used and distributed together with GStreamer and BMPx. This
//  permission is above and beyond the permissions granted by the GPL license
//  BMPx is covered by.

#include <config.h>

#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <iostream>
#include <istream>
#include <ostream>
#include <fstream>
#include <boost/format.hpp>

#include <unistd.h>
#include <errno.h>

#include "util.hpp"
#include "uri++.hpp"
#include "file_utils.hpp"

#include "x_play.hpp"

#ifndef __USE_BSD
# define __USE_BSD      /* Needed for mkdtemp () */
#endif

#include "main.hpp"

//////////////// Archive handling functions
namespace Bmp
{
  namespace Util
  {
    namespace
    {
      typedef std::list<Glib::ustring> DirEntries;

      enum ArchiveType
      {
          ARCHIVE_UNKNOWN = 0,
          ARCHIVE_DIR,
          ARCHIVE_TAR,
          ARCHIVE_TGZ,
          ARCHIVE_ZIP,
          ARCHIVE_TBZ2
      };

      struct ArchiveExtensionType
      {
          ArchiveType type;
          const gchar *ext;
      };

      typedef Glib::ustring (*ArchiveExtractFunc) (const Glib::ustring &,
                                                 const Glib::ustring &);

      ArchiveExtensionType archive_extensions[] =
      {
          {ARCHIVE_TAR, "tar"},
          {ARCHIVE_ZIP, "wsz"},
          {ARCHIVE_ZIP, "zip"},
          {ARCHIVE_TGZ, "tar.gz"},
          {ARCHIVE_TGZ, "tgz"},
          {ARCHIVE_TBZ2, "tar.bz2"},
          {ARCHIVE_TBZ2, "bz2"},
          {ARCHIVE_UNKNOWN, NULL}
      };

      Glib::ustring archive_extract_tar (const Glib::ustring &archive,
                                       const Glib::ustring &dest);

      Glib::ustring archive_extract_zip  (const Glib::ustring &archive,
                                        const Glib::ustring &dest);

      Glib::ustring archive_extract_tgz  (const Glib::ustring &archive,
                                        const Glib::ustring &dest);

      Glib::ustring archive_extract_tbz2 (const Glib::ustring &archive,
                                        const Glib::ustring &dest);

      ArchiveExtractFunc archive_extract_funcs[] =
      {
          0,
          0,
          archive_extract_tar,
          archive_extract_tgz,
          archive_extract_zip,
          archive_extract_tbz2
      };

      Glib::ustring
      escape_shell_chars (const Glib::ustring &string)
      {
	  char *out_c = g_shell_quote (string.c_str());
	  Glib::ustring out (out_c);
	  free (out_c);
	  return out;
      }

      Glib::ustring
      archive_extract_tar (const Glib::ustring &archive,
                           const Glib::ustring &dest)
      {
          static boost::format format ("%s >/dev/null xf %s -C %s");

          return (format % "tar" % archive % dest).str ();
      }

      Glib::ustring
      archive_extract_zip (const Glib::ustring &archive,
                           const Glib::ustring &dest)
      {
          static boost::format format ("%s >/dev/null -o -j %s -d %s");

          return (format % "unzip" % archive % dest).str ();
      }

      Glib::ustring
      archive_extract_tgz (const Glib::ustring &archive,
                           const Glib::ustring &dest)
      {
          static boost::format format ("%s >/dev/null xzf %s -C %s");

          return (format % "tar" % archive % dest).str ();
      }

      Glib::ustring
      archive_extract_tbz2 (const Glib::ustring &archive,
                            const Glib::ustring &dest)
      {
          static boost::format format ("bzip2 -dc %s | %s >/dev/null xf - -C %s");

          return (format % archive % "tar" % dest).str ();
      }

      ArchiveType
      archive_get_type (const Glib::ustring &filename)
      {
          gint i = 0;

          if (Glib::file_test (filename, Glib::FILE_TEST_IS_DIR))
              return ARCHIVE_DIR;

          while (archive_extensions[i].ext)
          {
              if (g_str_has_suffix (filename.c_str (), archive_extensions[i].ext))
              {
                  return archive_extensions[i].type;
              }
              i++;
          }

          return ARCHIVE_UNKNOWN;
      }

      bool
      find_file_actual (const Glib::ustring &dir_path,
                        const Glib::ustring &filename,
                        Glib::ustring       &found_path,
                        int                max_depth,
                        int                depth)
      {
          if (max_depth != -1 && depth > max_depth)
	    {
              return false;
	    }

          Glib::Dir dir (dir_path);
	  DirEntries entries (dir.begin(), dir.end());
	  dir.close ();

          for (DirEntries::const_iterator iter = entries.begin (),
					  end  = entries.end ();  iter != end;  ++iter)

	  {
	      Glib::ustring path = Glib::build_filename (dir_path, *iter);

	      if (path.length () > FILENAME_MAX)
	      {
		  g_warning (G_STRLOC ": ignoring path: name too long (%s)",
			     Glib::filename_to_utf8 (path).c_str ());
		  continue;
	      }

	      if (Glib::file_test (path, Glib::FILE_TEST_IS_REGULAR))
	      {
		  if (Glib::filename_to_utf8 (*iter).lowercase () ==
		      Glib::filename_to_utf8 (filename).lowercase ())
		  {
		      found_path = path;
		      return true;
		  }
	      }
	      else
	      if (Glib::file_test (path, Glib::FILE_TEST_IS_DIR))
	      {
		  bool result = find_file_actual (path, filename, found_path, max_depth, depth + 1);
		  if (result) return true;
	      }
	  }
	  return false;
      }

    } // <anonymous> namespace

    // FIXME: Return a Glib::ustring
    char *
    create_glob (const Glib::ustring &suffix)
    {
        GString *out;
        gint n = 0;

        out = g_string_new ("");

        while (suffix[n])
        {
            g_string_append_printf (out, "[%c%c]", g_ascii_tolower (suffix[n]), g_ascii_toupper (suffix[n]));
            n++;
        }

        return g_string_free (out, FALSE);
    }

    bool
    file_is_archive (const Glib::ustring &filename)
    {
        return (archive_get_type (filename) > ARCHIVE_DIR);
    }

    Glib::ustring
    archive_basename (const Glib::ustring &str_std)
    {
        gint i = 0;
        char *str = g_strdup (str_std.c_str ());

        while (archive_extensions[i].ext)
        {
            if (str_has_suffix_nocase (str, archive_extensions[i].ext))
            {
                const gchar *end = g_strrstr (str, archive_extensions[i].ext);
                if (end)
                {
                    /* The -1 is for the '.' in the extension */

                    //FIXME: Don't use C string copying here
                    char *result = g_strndup (str, end - str - 1);
                    Glib::ustring result_std (result);

                    g_free (result);
                    g_free (str);

                    return result_std;
                }
            }
            i++;
        }

        g_free (str);
        return Glib::ustring ("");
    }

    Glib::ustring
    archive_decompress (const Glib::ustring &filename)
    {
        Glib::ustring escaped_filename, command;

        ArchiveType type;
        if ((type = archive_get_type (filename)) <= ARCHIVE_DIR)
            return Glib::ustring ("");

        char *c_tmp_dir = g_strdup (Glib::build_filename (g_get_tmp_dir (), "bmp.XXXXXXXX").c_str ());

        if (!mkdtemp (c_tmp_dir))
        {
            g_message ("Unable to decompress archive '%s': could not create temporary directory: %s",
                       Glib::filename_to_utf8 (filename).c_str (), g_strerror (errno));

            return Glib::ustring ("");
        }

        Glib::ustring tmp_dir (c_tmp_dir);
        g_free (c_tmp_dir);

        escaped_filename = escape_shell_chars (filename);
        command = archive_extract_funcs[type] (escaped_filename, tmp_dir);

        if (!command.length ())
            return Glib::ustring ("");

        if (system (command.c_str ()))
        {
            g_message ("Unable to decompress archive '%s'", Glib::filename_to_utf8 (filename).c_str ());
            return Glib::ustring ("");
        }

        return tmp_dir;
    }

    bool
    find_file (const Glib::ustring &dir_path,
               const Glib::ustring &filename,
               Glib::ustring       &found_path,
               int                max_depth)
    {
        found_path = "";
        return find_file_actual (dir_path, filename, found_path, max_depth, 0);
    }

    // Deletes a directory recursively
    void
    del_directory (const Glib::ustring &dir_path)
    {
	Glib::Dir dir (dir_path);
	DirEntries entries (dir.begin(), dir.end());
	dir.close ();

        for (DirEntries::const_iterator iter = entries.begin (), end = entries.end ();
             iter != end;
             ++iter)
        {
            if (*iter != "." && *iter != "..")
            {
                Glib::ustring full_path = Glib::build_filename (dir_path, *iter);

                if (Glib::file_test (full_path, Glib::FILE_TEST_IS_DIR))
                    del_directory (full_path);
                else
                    unlink (full_path.c_str ());
            }
        }

        rmdir (dir_path.c_str ());
    }

    // "Collects" a path of audio files recursively
    void
    collect_path (const Glib::ustring &dir_path,
                  FileList          &collection,
                  bool               clear)
    {
        if (clear)
        {
            collection.clear ();
        }

	Glib::Dir dir (dir_path);
	DirEntries entries (dir.begin(), dir.end());
	dir.close ();

        for (DirEntries::const_iterator iter = entries.begin (), end = entries.end ();
             iter != end;
             ++iter)
        {
            if (*iter != "." && *iter != "..")
            {
                Glib::ustring full_path = Glib::build_filename (dir_path, *iter);

                if (Glib::file_test (full_path, Glib::FILE_TEST_IS_REGULAR))
                {
		    if (play->is_audio_file (*iter))
		      {
			Bmp::URI u (Glib::filename_to_uri (full_path));
			u.unescape ();
                        collection.push_back (Glib::ustring(u));
		      }
                }
                else if (Glib::file_test (full_path, Glib::FILE_TEST_IS_DIR))
                {
                    collect_path (full_path, collection, false);
                }
            }
        }
    }

    void
    dir_for_each_entry (const Glib::ustring &path,
                        DirForeachSlot	     slot)
    {
	Glib::Dir dir (path);
	DirEntries entries (dir.begin(), dir.end());
	dir.close ();

        for (DirEntries::const_iterator iter = entries.begin(), end = entries.end(); iter != end; ++iter)
        {
            Glib::ustring full_path = Glib::build_filename (path, *iter);
            if (slot (full_path)) break;
        }
    }

    void
    files_writable (const FileList& list,
		    FileList& non_writable)
    {

      for (FileList::const_iterator uri = list.begin (), end = list.end (); uri != end; ++uri)
      {
	Glib::ustring filename = Glib::filename_from_uri (*uri);

	struct stat info;
	if (stat (filename.c_str (), &info))
	{
	  non_writable.push_back (*uri);
	  continue;
	}

	//Get stat info 
	uid_t  stat_uid  = info.st_uid;
	gid_t  stat_gid  = info.st_gid;
	mode_t stat_mode = info.st_mode;

	//Check if world-writable
	if (stat_mode & S_IWOTH) continue;

	//Get the process uid/gid
	uid_t p_uid = geteuid ();	
	gid_t p_gid = getegid ();

	//Check whether group has write permission	
	if ((p_gid == stat_gid) && (stat_mode & S_IWGRP)) continue;

	//Check whether user has write permission
	if ((p_uid == stat_uid) && (stat_mode & S_IWUSR)) continue;

	non_writable.push_back (*uri);
      }
    }

    bool
    files_taggable (const FileList& list)
    {
	char* const suffixes[] = { 
	    "mp3",
	    "ogg",
	    "mpc",
	    NULL,
	};

	
	for (FileList::const_iterator uri = list.begin (), end = list.end (); uri != end; ++uri)
	{
	    if (!str_has_suffixes_nocase ((*uri).c_str(), suffixes)) return false;
	}

	return true;
    }

    void
    copy_file (const gchar  *source,
	       const gchar  *dest)
    {
	using namespace std;

	ifstream infile	  (source, ios::binary);
	ofstream outfile  (dest, ios::binary);
	outfile << infile.rdbuf();
	outfile.close ();
	infile.close ();
    }

  } // namespace Util

} // namespace Bmp

/////////////////////////////////////////////////////////////////////


